# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-10-17 13:29
from __future__ import unicode_literals

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations
from django.db import models

import orchestra.models.core.mixins
import orchestra.utils.models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0069_update_currencies'),
    ]

    operations = [
        migrations.CreateModel(
            name='Todo',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now)),
                ('is_deleted', models.BooleanField(default=False)),
                ('description', models.CharField(max_length=200)),
                ('completed', models.BooleanField(default=False)),
                ('task', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='todos', to='orchestra.Task')),
            ],
            bases=(orchestra.models.core.mixins.TodoMixin, orchestra.utils.models.DeleteMixin, models.Model),
        ),
        migrations.AlterField(
            model_name='taskassignment',
            name='worker',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='assignments', to='orchestra.Worker'),
        ),
    ]



import json
from unittest.mock import patch

from django.test import override_settings
from django.test import TestCase
from django.test import TransactionTestCase

from orchestra.tests.helpers.notifications import MockMail
from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.communication.tests.helpers.slack import MockSlacker
from orchestra.utils.load_json import load_encoded_json

# Don't log logger errors.
import logging
logging.disable(logging.CRITICAL)


class OrchestraTestHelpersMixin(object):

    def setUp(self):
        super().setUp()
        # Rename the django Client to request_client
        self.request_client = self.client
        # maxDiff prevents the test runner from suppressing the diffs on
        # assertEqual, which is nice when you have large string comparisons as
        # we do in this test to assert the expected JSON blob responses.
        self.maxDiff = None

        # Without patching the slack API calls, the tests hang indefinitely
        # and you'll need to restart your boot2docker.
        self.slack = MockSlacker()
        patcher = patch(
            'orchestra.communication.slack.Slacker',
            return_value=self.slack
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        self.addCleanup(self.slack.clear)

        # Though mail isn't sent in django tests, we mock out send_mail here to
        # have access to sent messages for testing.
        self.mail = MockMail()
        patcher = patch('orchestra.utils.notifications.send_mail',
                        side_effect=self.mail.send_mail)
        patcher.start()
        self.addCleanup(patcher.stop)
        self.addCleanup(self.mail.clear)

    def ensure_response(self,
                        response,
                        expected_json_payload,
                        expected_status_code):
        self.assertEqual(response.status_code, expected_status_code)
        returned = load_encoded_json(response.content)
        self.assertEqual(returned,
                         expected_json_payload)

    def _submit_assignment(self, request_client, task_id, data=None,
                           seconds=1, command='submit'):
        if data is None:
            data = {'test': 'test'}
        request = json.dumps(
            {'task_id': task_id, 'task_data': data, 'command_type': command})

        return request_client.post(
            '/orchestra/api/interface/submit_task_assignment/',
            request,
            content_type='application/json')

    def assertModelInstanceExists(self,
                                  model_class,
                                  lookup_attributes,
                                  check_attributes={}):
        try:
            model_instance = model_class.objects.get(**lookup_attributes)
        except model_class.DoesNotExist:
            self.fail('{} instance not created in database.'
                      .format(str(model_class)))
        for attribute_name, expected_value in check_attributes.items():
            self.assertEqual(getattr(model_instance, attribute_name),
                             expected_value)
        return model_instance

    def assertModelInstanceNotExists(self, model_class, lookup_attributes):
        with self.assertRaises(model_class.DoesNotExist):
            model_class.objects.get(**lookup_attributes)


class AuthenticatedUserMixin(object):

    def authenticate_user(self):
        password = 'test'
        auth_user = UserFactory(
            username='test_user',
            email='test_user@orchestra.com',
            first_name='first name',
            last_name='last name',
            is_active=True,
            password=password)
        self.assertTrue(self.request_client.login(
            username=auth_user.username, password=password))
        return auth_user


@override_settings(ORCHESTRA_SLACK_STAFFBOT_TOKEN='test-token')
class OrchestraTestCase(OrchestraTestHelpersMixin, TestCase):
    # NOTE(lydia): Mixin should go before TestCase because when mixin calls
    # super().setUp(), it looks to the base of mixin (which is Object - doesn't
    # have setUp() defined, then to the right for setUp() (which is defined
    # in TestCase)
    pass


@override_settings(ORCHESTRA_SLACK_STAFFBOT_TOKEN='test-token')
class OrchestraTransactionTestCase(OrchestraTestHelpersMixin,
                                   TransactionTestCase):
    # NOTE(lydia): See note above about multiple inheritance ordering.
    pass


class OrchestraAuthenticatedTestCase(OrchestraTestCase,
                                     AuthenticatedUserMixin):
    # NOTE(lydia): See note above about multiple inheritance ordering.
    pass


class OrchestraModelTestCase(OrchestraTestCase):
    """
    This is a basic test to ensure that we update the __str__ method on models
    when we modify fields. It is a simple way to catch errors before we break
    the admin page because of model changes.

    NOTE: This does not test the validity of __str__ it just verifies that the
    function can run without error.
    """
    __test__ = False
    model = None
    model_kwargs = {}

    def test_to_string(self):
        instance = self.model(**self.model_kwargs)
        self.assertEqual(str(instance), str(instance))


class EndpointTestCase(OrchestraTestCase):

    def _verify_bad_request(self, response, message):
        self.assertEqual(response.status_code, 400)
        data = load_encoded_json(response.content)
        self.assertEqual(data['message'], message)
        self.assertEqual(data['error'], 400)



# Generated by Django 2.2.9 on 2020-04-09 14:27

from django.db import migrations
import jsonfield.fields


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0083_update_staffbot_request_status_choices'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowversion',
            name='abort_completion_function',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



import datetime
from unittest.mock import patch

from django.utils import timezone

from orchestra.core.errors import TimerError
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import TaskTimer
from orchestra.models import TimeEntry
from orchestra.models import Worker
from orchestra.tests.helpers import OrchestraTransactionTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.time_tracking import get_timer_current_duration
from orchestra.utils.time_tracking import get_timer_object
from orchestra.utils.time_tracking import start_timer
from orchestra.utils.time_tracking import stop_timer


class TaskTimerTests(OrchestraTransactionTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = Worker.objects.get(user__username='test_user_6')
        self.tasks = Task.objects.filter(assignments__worker=self.worker)
        self.task = self.tasks[0]
        self.assignment = (self.task.assignments.filter(worker=self.worker)
                           .first())
        self.time_entry_data = {'date': '2016-05-01',
                                'time_worked': '00:30:00',
                                'description': 'test description'}
        self.time = timezone.now()

        def fake_now():
            """ Patch timezone.now so we can test timer. """
            return self.time

        patcher = patch('orchestra.utils.time_tracking.timezone.now', fake_now)
        self.addCleanup(patcher.stop)
        patcher.start()

    def test_get_timer_object_created(self):
        with self.assertRaises(TaskTimer.DoesNotExist):
            self.worker.timer
        timer = get_timer_object(self.worker)
        self.assertEqual(timer, self.worker.timer)

    def test_get_timer_object_not_created(self):
        timer = TaskTimer(worker=self.worker)
        timer.save()
        new_timer = get_timer_object(self.worker)

        # Function should return the already created timer object.
        self.assertEqual(new_timer, timer)

    def test_start_timer(self):
        start_timer(self.worker)
        timer = self.worker.timer
        self.assertEqual(timer.start_time, self.time)

    def test_start_timer_already_running(self):
        timer = TaskTimer(worker=self.worker,
                          start_time=self.time)
        timer.save()
        with self.assertRaisesRegex(TimerError, 'Timer has already started'):
            start_timer(self.worker)

    def test_start_timer_with_assignment(self):
        start_timer(self.worker, assignment_id=self.assignment.id)
        timer = self.worker.timer
        self.assertEqual(timer.assignment, self.assignment)

    def test_start_timer_worker_not_assigned(self):
        # Find task assignment that worker is not assigned to.
        assignment = (TaskAssignment.objects.exclude(worker=self.worker)
                      .first())
        with self.assertRaises(TaskAssignment.DoesNotExist):
            start_timer(self.worker, assignment_id=assignment.id)

    def test_stop_timer(self):
        timer = TaskTimer(worker=self.worker,
                          assignment=self.assignment,
                          start_time=self.time)
        timer.save()
        start_time = self.time
        self.time = self.time + datetime.timedelta(hours=1)
        time_entry = stop_timer(self.worker)

        # Time entry object is created correctly.
        self.assertEqual(time_entry.worker, self.worker)
        self.assertEqual(time_entry.date, start_time.date())
        self.assertEqual(time_entry.time_worked, datetime.timedelta(hours=1))
        self.assertEqual(time_entry.assignment, self.assignment)
        self.assertEqual(time_entry.timer_start_time, start_time)
        self.assertEqual(time_entry.timer_stop_time, self.time)

        # Timer object should be reset.
        timer.refresh_from_db()
        self.assertIsNone(timer.start_time)
        self.assertIsNone(timer.stop_time)

    def test_stop_timer_not_running(self):
        timer = TaskTimer(worker=self.worker)
        timer.save()
        with self.assertRaisesRegex(TimerError, 'Timer not started'):
            stop_timer(self.worker)

    @patch('orchestra.utils.time_tracking._reset_timer', side_effect=Exception)
    def test_stop_timer_atomic(self, mock_reset):
        timer = TaskTimer(worker=self.worker, start_time=self.time)
        timer.save()
        time_entries_count = TimeEntry.objects.count()
        self.assertIsNone(timer.stop_time)

        with self.assertRaises(Exception):
            stop_timer(self.worker)

        # Verify that transaction rolled back.
        timer.refresh_from_db()
        self.assertIsNone(timer.stop_time)
        self.assertEqual(TimeEntry.objects.count(), time_entries_count)

    def test_get_timer_current_duration(self):
        timer = TaskTimer(worker=self.worker,
                          assignment=self.assignment,
                          start_time=self.time)
        timer.save()
        self.time = self.time + datetime.timedelta(hours=1)
        duration = get_timer_current_duration(self.worker)
        self.assertEqual(duration, datetime.timedelta(hours=1))

    def test_get_timer_current_duration_no_start_time(self):
        timer = TaskTimer(worker=self.worker,
                          assignment=self.assignment)
        timer.save()
        duration = get_timer_current_duration(self.worker)
        self.assertIsNone(duration)



import logging
import re

from django.conf import settings
from django.urls import reverse
from django.db.models import Q
from django.template.loader import render_to_string

from orchestra.bots.basebot import BaseBot
from orchestra.bots.errors import SlackUserUnauthorized
from orchestra.communication.errors import SlackError
from orchestra.communication.mail import html_from_plaintext
from orchestra.communication.mail import send_mail
from orchestra.communication.slack import format_slack_message
from orchestra.communication.utils import close_open_staffbot_requests
from orchestra.core.errors import TaskAssignmentError
from orchestra.core.errors import TaskStatusError
from orchestra.models import CommunicationPreference
from orchestra.models import StaffBotRequest
from orchestra.models import StaffingRequestInquiry
from orchestra.models import Task
from orchestra.models import TaskAssignment
from orchestra.models import Worker
from orchestra.utils.notifications import message_experts_slack_group
from orchestra.utils.task_lifecycle import assert_new_task_status_valid
from orchestra.utils.task_lifecycle import role_counter_required_for_new_task

logger = logging.getLogger(__name__)


# Types of responses we can send to slack
VALID_RESPONSE_TYPES = {'ephemeral', 'in_channel'}

# A regex for markdown links, which are of the form [link text](url).
# We capture link text and url as an text that isn't a closing ] or )
# respectively.
MARKDOWN_LINK_REGEX = re.compile('\[(?P<text>[^\]]*)\]\((?P<url>[^\)]*)\)')


class StaffBot(BaseBot):

    commands = (
        (r'staff\s+(?P<task_id>[0-9]+)', 'staff'),
        (r'restaff\s+(?P<task_id>[0-9]+)\s+(?P<username>[\w.@+-]+)',
         'restaff'),
    )
    task_does_not_exist_error = 'Task {} does not exist'
    task_assignment_error = 'Task {} got an error: "{}"'
    worker_does_not_exist = 'Worker with username {} does not exist'
    nonunique_slack_id = 'There are multiple workers with slack_user_id {}'
    staffing_is_not_allowed = (
        'Staffing of task {} is not allowed at this state')
    task_assignment_does_not_exist_error = (
        'TaskAssignment associated with user {} and task {} does not exist.')
    not_authorized_error = 'You are not authorized to staff projects!'
    staffing_success = 'Got it! I will start staffing task {}!'
    restaffing_success = 'Got it! I will start restaffing task {}!'

    def __init__(self, **kwargs):
        default_config = getattr(settings, 'STAFFBOT_CONFIG', {})
        default_config.update(kwargs)
        token = settings.ORCHESTRA_SLACK_STAFFBOT_TOKEN
        super().__init__(token, **default_config)

    def help(self):
        return format_slack_message(
            'Use `/staffbot` to staff or restaff a Task within Orchestra.',
            attachments=[
                {
                    'pretext': '`/staffbot staff <task-id>`',
                    'text': ('Ask qualified experts to work on'
                             ' the given `<task-id>`.'),
                    'mrkdwn_in': ['text', 'pretext'],
                },
                {
                    'pretext': '`/staffbot restaff <task-id> <username>`',
                    'text': ('Remove the given `<username>` from'
                             ' the task and find another qualified expert'),
                    'mrkdwn_in': ['text', 'pretext'],
                }
            ])

    def staff(self, task_id,
              request_cause=StaffBotRequest.RequestCause.USER.value):
        """
        This function handles staffing a request for the given task_id.
        """
        command = 'staff {}'.format(task_id)
        try:
            task = Task.objects.get(id=task_id)
            required_role_counter = role_counter_required_for_new_task(task)
            error_msg = None
            assert_new_task_status_valid(task.status)
        except TaskStatusError:
            error_msg = self.staffing_is_not_allowed.format(task_id)
        except Task.DoesNotExist:
            error_msg = self.task_does_not_exist_error.format(task_id)
        except TaskAssignmentError as error:
            error_msg = self.task_assignment_error.format(task_id, error)

        if error_msg is not None:
            logger.exception(error_msg)
            return format_slack_message(
                command,
                attachments=[{
                    'color': 'danger',
                    'title': 'Error',
                    'text': error_msg
                }])

        close_open_staffbot_requests(task)
        StaffBotRequest.objects.create(
            task=task,
            required_role_counter=required_role_counter,
            request_cause=request_cause)

        slack_message = self.staffing_success.format(task_id)
        message_experts_slack_group(task.project.slack_group_id, slack_message)
        return format_slack_message(
            command,
            attachments=[{
                'color': 'good',
                'title': 'Success',
                'text': slack_message
            }])

    def restaff(self, task_id, username,
                request_cause=StaffBotRequest.RequestCause.USER.value):
        """
        This function handles restaffing a request for the given task_id.
        The current user for the given username is removed, and a new user
        is found.
        """
        command = 'restaff {} {}'.format(task_id, username)
        try:
            error_msg = None

            worker = Worker.objects.filter(
                Q(user__username=username) | Q(slack_username=username))

            if worker.exists():
                worker = worker.first()
            else:
                error_msg = self.worker_does_not_exist.format(username)
                return format_slack_message(
                    command,
                    attachments=[{
                        'color': 'danger',
                        'title': 'Error',
                        'text': error_msg
                    }])
            task = Task.objects.get(id=task_id)
            task_assignment = TaskAssignment.objects.get(worker=worker,
                                                         task=task)
            required_role_counter = task_assignment.assignment_counter

        except Task.DoesNotExist:
            error_msg = self.task_does_not_exist_error.format(task_id)
        except TaskAssignment.DoesNotExist:
            error_msg = (self.task_assignment_does_not_exist_error
                         .format(username, task_id))
        except TaskAssignmentError as error:
            error_msg = self.task_assignment_error.format(task_id, error)

        if error_msg is not None:
            logger.exception(error_msg)
            return format_slack_message(
                command,
                attachments=[{
                    'color': 'danger',
                    'title': 'Error',
                    'text': error_msg
                }])

        close_open_staffbot_requests(task)
        StaffBotRequest.objects.create(
            task=task,
            required_role_counter=required_role_counter,
            request_cause=request_cause)
        slack_message = self.restaffing_success.format(task_id)

        message_experts_slack_group(task.project.slack_group_id, slack_message)
        return format_slack_message(
            command,
            attachments=[{
                'color': 'good',
                'title': 'Success',
                'text': slack_message
            }])

    def send_worker_tasks_available_reminder(self, worker):
        communication_type = (
            CommunicationPreference.CommunicationType.NEW_TASK_AVAILABLE.value)

        communication_preference = CommunicationPreference.objects.get(
            communication_type=communication_type,
            worker=worker)

        user = communication_preference.worker.user
        tasks_url = self._get_staffing_url(
            'orchestra:communication:available_staffing_requests',
            None)
        context = {'user': user,
                   'tasks_url': tasks_url}
        if communication_preference.can_email():
            message_body = render_to_string(
                'communication/tasks_available_email.txt',
                context)
            email = communication_preference.worker.user.email
            self._send_staffing_request_by_mail(email, message_body)

        if communication_preference.can_slack():
            message_body = render_to_string(
                'communication/tasks_available_slack.txt',
                context)
            self._send_staffing_request_by_slack(
                communication_preference.worker, message_body)

    def send_task_to_worker(self, worker, staffbot_request):
        """
        Send the task to the worker for them to accept or reject.
        """
        communication_type = (
            CommunicationPreference.CommunicationType.NEW_TASK_AVAILABLE.value)

        communication_preference = CommunicationPreference.objects.get(
            communication_type=communication_type,
            worker=worker)

        if communication_preference.can_email():
            email_method = (
                StaffingRequestInquiry.CommunicationMethod.EMAIL.value)
            staffing_request_inquiry = StaffingRequestInquiry.objects.create(
                communication_preference=communication_preference,
                communication_method=email_method,
                request=staffbot_request)
            message = self._get_staffing_request_message(
                staffing_request_inquiry,
                'communication/new_task_available_email.txt')
            email = communication_preference.worker.user.email
            self._send_staffing_request_by_mail(email, message)

        if communication_preference.can_slack():
            slack_method = (
                StaffingRequestInquiry.CommunicationMethod.SLACK.value)
            staffing_request_inquiry = StaffingRequestInquiry.objects.create(
                communication_preference=communication_preference,
                communication_method=slack_method,
                request=staffbot_request)
            message = self._get_staffing_request_message(
                staffing_request_inquiry,
                'communication/new_task_available_slack.txt')
            self._send_staffing_request_by_slack(
                communication_preference.worker, message)

    def _get_staffing_url(self, reverse_string, url_kwargs):
        return '{}{}'.format(
            settings.ORCHESTRA_URL,
            reverse(reverse_string, kwargs=url_kwargs))

    def get_staffing_request_metadata(self, staffing_request_inquiry):
        url_kwargs = {
            'staffing_request_inquiry_id': staffing_request_inquiry.pk
        }
        accept_url = self._get_staffing_url(
            'orchestra:communication:accept_staffing_request_inquiry',
            url_kwargs)
        reject_url = self._get_staffing_url(
            'orchestra:communication:reject_staffing_request_inquiry',
            url_kwargs)

        # TODO(joshblum): handle urls if present in the detailed_description to
        # convert for slack
        staffbot_request = staffing_request_inquiry.request
        detailed_description = (
            staffbot_request.task.get_detailed_description()
        )
        workflow_description = (
            staffbot_request.task.project
            .workflow_version.workflow.description
        )
        project_description = (
            staffbot_request.task.project
            .short_description
        )
        step_description = (
            staffbot_request.task.step.description)
        user = (staffing_request_inquiry.communication_preference.
                worker.user)
        return {
            'user': user,
            'accept_url': accept_url,
            'reject_url': reject_url,
            'role_counter': staffbot_request.required_role_counter,
            'step_description': step_description,
            'workflow_description': workflow_description,
            'project_description': project_description,
            'detailed_description': detailed_description,
            'task': staffbot_request.task
        }

    def _get_staffing_request_message(self, staffing_request_inquiry,
                                      template):
        context = self.get_staffing_request_metadata(staffing_request_inquiry)
        context['available_requests_url'] = self._get_staffing_url(
            'orchestra:communication:available_staffing_requests', {})
        message_body = render_to_string(template, context)
        return message_body

    def _send_staffing_request_by_mail(self, email, message):
        html_message = html_from_plaintext(message)
        # Slack does not accept html tags, so we want to let markdown add some
        # simple things like <p>
        send_mail('A new task is available for you',
                  message,
                  settings.ORCHESTRA_NOTIFICATIONS_FROM_EMAIL,
                  [email],
                  html_message=html_message)

    def _send_staffing_request_by_slack(self, worker, message):
        if worker.slack_user_id is None:
            logger.warning('Worker {} does not have a slack id'.format(
                worker))
            return
        try:
            # Replace any Markdown-style links with Slack-style links.
            message = MARKDOWN_LINK_REGEX.sub(r'<\g<url>|\g<text>>', message)
            self.slack.chat.post_message(worker.slack_user_id, message)
        except SlackError:
            logger.warning('Invalid slack id {} {}'.format(
                worker.slack_user_id, message))

    def validate(self, data):
        """
        Handle a request received from slack. First we validate the
        request and then pass the message to the appropriate handler.
        """
        slack_user_id = data.get('user_id')
        username = data.get('user_name')

        workers = Worker.objects.filter(slack_user_id=slack_user_id)
        worker = workers.first()
        if workers.count() > 1:
            raise SlackUserUnauthorized(
                self.nonunique_slack_id.format(slack_user_id))
        elif worker is None:
            raise SlackUserUnauthorized(
                'Worker {} not found. slack_user_id: {}'.format(
                    username, slack_user_id))
        elif not worker.is_project_admin():
            raise SlackUserUnauthorized(self.not_authorized_error)
        data = super().validate(data)
        return data



from django.test import TestCase
from django.utils import timezone

from orchestra.tests.helpers.fixtures import UserFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.utils.common_helpers import get_update_message


class ViewHelpersTests(TestCase):
    def setUp(self):
        super().setUp()
        project = ProjectFactory()
        step = StepFactory()
        self.old_title = 'Old title'
        self.new_title = 'New title'
        self.old_details = 'Old details'
        self.new_details = 'New details'
        self.old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            completed=False,
            project=project,
            step=step)
        self.new_todo = TodoFactory(
            title=self.new_title,
            details=self.new_details,
            completed=True,
            project=project,
            step=step)
        self.sender = UserFactory()

    def test_empty_msg_if_no_updates(self):
        old_todo = self.new_todo
        msg = get_update_message(old_todo, self.new_todo, self.sender)
        expected_msg = ''
        self.assertEqual(msg, expected_msg)

    def test_fields_updated_completed_with_sender(self):
        msg = get_update_message(self.old_todo, self.new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: marked complete, '
            'changed title and details'
            ).format(self.sender.username, self.new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_updated_completed_without_sender(self):
        msg = get_update_message(self.old_todo, self.new_todo)
        expected_msg = (
            '`{}` has been updated: marked complete, '
            'changed title and details'
            ).format(self.new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_not_updated_completed(self):
        new_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            completed=True)
        msg = get_update_message(self.old_todo, new_todo, self.sender)
        expected_msg = '{} has updated `{}`: marked complete'.format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_not_updated_not_relevant(self):
        new_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details,
            completed=False,
            skipped_datetime=timezone.now())
        msg = get_update_message(self.old_todo, new_todo, self.sender)
        expected_msg = '{} has updated `{}`: marked not relevant'.format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

        # No sender
        msg = get_update_message(self.old_todo, new_todo)
        expected_msg = '`{}` has been updated: marked not relevant'.format(
            new_todo.title)
        self.assertEqual(msg, expected_msg)

    def test_fields_updated(self):
        old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            details=self.new_details)
        msg = get_update_message(old_todo, new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: changed title and details').format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)

        # Only title has changed
        old_todo = TodoFactory(
            title=self.old_title,
            details=self.old_details)
        new_todo = TodoFactory(
            title=self.new_title,
            details=self.old_details)
        msg = get_update_message(old_todo, new_todo, self.sender)
        expected_msg = (
            '{} has updated `{}`: changed title').format(
            self.sender.username, new_todo.title)
        self.assertEqual(msg, expected_msg)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import jsonfield.fields
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Certification',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('slug', models.CharField(unique=True, max_length=200)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('required_certifications', models.ManyToManyField(
                    related_name='required_certifications_rel_+', to='orchestra.Certification')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Process',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('slug', models.CharField(unique=True, max_length=200)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('start_datetime', models.DateTimeField(auto_now_add=True)),
                ('priority', models.IntegerField()),
                ('task_class', models.IntegerField(
                    choices=[(0, b'Training tasks'), (1, b'A real task')])),
                ('process', models.ForeignKey(
                    on_delete=models.CASCADE, to='orchestra.Process')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Step',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('slug', models.CharField(max_length=200)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('worker_type', models.IntegerField(
                    choices=[(0, b'Human'), (1, b'Machine')])),
                ('review_policy', jsonfield.fields.JSONField()),
                ('user_interface', jsonfield.fields.JSONField()),
                ('depends_on', models.ManyToManyField(
                    related_name='depends_on_rel_+', to='orchestra.Step')),
                ('process', models.ForeignKey(
                    on_delete=models.CASCADE, to='orchestra.Process')),
                ('required_certifications', models.ManyToManyField(
                    to='orchestra.Certification')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('status', models.IntegerField(choices=[(0, b'Awaiting Processing'), (1, b'Processing'), (
                    2, b'Pending Review'), (3, b'Reviewing'), (4, b'Post-review Processing'), (5, b'Complete')])),
                ('step', models.ForeignKey(
                    on_delete=models.CASCADE, to='orchestra.Step')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TaskAssignment',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('start_datetime', models.DateTimeField(auto_now_add=True)),
                ('status', models.IntegerField(
                    choices=[(0, b'Procesing'), (1, b'Submitted')])),
                ('in_progress_task_data', jsonfield.fields.JSONField()),
                ('snapshots', jsonfield.fields.JSONField()),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='Worker',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('start_datetime', models.DateTimeField(auto_now_add=True)),
                ('user', models.ForeignKey(
                    on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='WorkerCertification',
            fields=[
                ('id', models.AutoField(
                    verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('task_class', models.IntegerField(
                    choices=[(0, b'Training tasks'), (1, b'A real task')])),
                ('role', models.IntegerField(
                    choices=[(0, b'Entry-level'), (1, b'Reviewer')])),
                ('certification', models.ForeignKey(on_delete=models.CASCADE,
                                                    to='orchestra.Certification')),
                ('worker', models.ForeignKey(
                    on_delete=models.CASCADE, to='orchestra.Worker')),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name='taskassignment',
            name='worker',
            field=models.ForeignKey(
                on_delete=models.CASCADE, to='orchestra.Worker'),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='step',
            unique_together=set([('process', 'slug')]),
        ),
    ]



from unittest.mock import patch

from django.urls import reverse
from django.test import Client as RequestClient
from django.test import override_settings

from orchestra.bots.staffbot import StaffBot
from orchestra.bots.tests.fixtures import get_mock_slack_data
from orchestra.models import Task
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers.fixtures import TaskFactory
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.task_lifecycle import assign_task


class StaffBotViewTest(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.worker = self.workers[0]
        self.worker.user.is_superuser = True
        self.worker.user.save()

        self.request_client = RequestClient(username=self.worker.user.username,
                                            password='defaultpassword')
        self.url = reverse('orchestra:bots:staffbot')
        self.staffbot = StaffBot()

    def assert_response(self, response, default_error_text=None):
        self.assertEqual(response.status_code, 200)
        data = load_encoded_json(response.content)
        if default_error_text is not None:
            self.assertTrue(default_error_text in data.get('text', ''))

    def test_unauthorized_user(self):
        worker1 = self.workers[1]
        request_client = RequestClient(username=worker1.user.username,
                                       password='defaultpassword')
        data = get_mock_slack_data(
            user_id=worker1.slack_user_id)
        response = request_client.post(self.url, data)
        self.assert_response(response,
                             default_error_text=StaffBot.not_authorized_error)

        data['user_id'] = 'fake_id'
        response = request_client.post(self.url, data)
        self.assert_response(response,
                             default_error_text='not found')

    def test_get_not_allowed(self):
        response = self.request_client.get(self.url)
        self.assertEqual(response.status_code, 405)

    def test_post_valid_data(self):
        data = get_mock_slack_data(
            user_id=self.worker.slack_user_id)
        response = self.request_client.post(self.url, data)
        self.assert_response(response)

    @override_settings(ORCHESTRA_SLACK_STAFFBOT_TOKEN='')
    def test_post_invalid_token(self):
        data = get_mock_slack_data(
            user_id=self.worker.slack_user_id)
        response = self.request_client.post(self.url, data)
        self.assert_response(
            response, default_error_text='Invalid token')

    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_mail')
    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_slack')
    def test_staff_command(self, mock_slack, mock_mail):
        task = TaskFactory(status=Task.Status.AWAITING_PROCESSING)
        data = get_mock_slack_data(
            text='staff {}'.format(task.id),
            user_id=self.worker.slack_user_id)
        response = self.request_client.post(self.url, data)
        self.assertEqual(
            load_encoded_json(response.content)['attachments'][0]['text'],
            self.staffbot.staffing_success.format(task.id))

        task = TaskFactory(status=Task.Status.PENDING_REVIEW)
        data = get_mock_slack_data(
            text='staff {}'.format(task.id),
            user_id=self.worker.slack_user_id)

        response = self.request_client.post(self.url, data)
        self.assertEqual(
            load_encoded_json(response.content)['attachments'][0]['text'],
            self.staffbot.staffing_success.format(task.id))

    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_mail')
    @patch('orchestra.bots.staffbot.StaffBot._send_staffing_request_by_slack')
    def test_restaff_command(self, mock_slack, mock_mail):
        data = get_mock_slack_data(user_id=self.worker.slack_user_id)

        task = (
            Task.objects.filter(status=Task.Status.AWAITING_PROCESSING)
            .first())
        worker = self.workers[0]
        task = assign_task(worker.id, task.id)
        command = 'restaff {} {}'.format(task.id, worker.user.username)
        data = get_mock_slack_data(
            text=command,
            user_id=self.worker.slack_user_id)

        response = self.request_client.post(self.url, data)
        self.assertEqual(
            load_encoded_json(response.content)['attachments'][0]['text'],
            self.staffbot.restaffing_success.format(task.id))



# -*- coding: utf-8 -*-
# Generated by Django 1.9.13 on 2017-10-25 09:08
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0072_migrate_staffing_priority'),
    ]

    operations = [
        migrations.RemoveField(  # manually-reviewed
            model_name='worker',
            name='staffing_priority',
        ),
    ]



import csv
import json

from django.conf import settings
from django.db import transaction
from django.utils import timezone
from tempfile import NamedTemporaryFile

from orchestra.core.errors import TodoListTemplateValidationError
from orchestra.google_apps.permissions import write_with_link_permission
from orchestra.google_apps.service import Service
from orchestra.google_apps.convenience import get_google_spreadsheet_as_csv
from orchestra.models import TodoListTemplateImportRecord


REMOVE_IF_HEADER = 'Remove if'
SKIP_IF_HEADER = 'Skip if'


def _write_template_rows(writer, todo, depth):
    """Recursively emits the rows of a to-do list template to the
    CSV `writer`. The format of each row is described in
    `export_to_spreadsheet` below. The `depth` of this row
    dictates the number of empty cells by which to indent
    this row's description.
    """
    writer.writerow(
        [json.dumps(todo.get('remove_if', [])),
         json.dumps(todo.get('skip_if', []))] +
        ([''] * depth) +
        [todo.get('description', '')])
    # `reversed` iteration because the JSON-serialized order of
    # children is the opposite of the top-to-bottom order in the
    # spreadsheet.
    for child in reversed(todo.get('items', [])):
        _write_template_rows(writer, child, depth + 1)


def _upload_csv_to_google(spreadsheet_name, file):
    service = Service(settings.GOOGLE_P12_PATH,
                      settings.GOOGLE_SERVICE_EMAIL)
    sheet = service.insert_file(
        spreadsheet_name,
        '',
        settings.ORCHESTRA_TODO_LIST_TEMPLATE_EXPORT_GDRIVE_FOLDER,
        'text/csv',
        'application/vnd.google-apps.spreadsheet',
        file.name
    )
    service.add_permission(sheet['id'], write_with_link_permission)
    return sheet['alternateLink']


def export_to_spreadsheet(todo_list_template):
    """
    Recursively descend down the to-do tree of a `todo_list_template`
    and add a row per to-do to a CSV, uploading it to a Google Sheet
    when done. The first two columns of the sheet capture json-encoded
    remove_if and skip_if logic. After that, the text in the `i`'th
    column represents a to-do's title that is `i` levels nested in the
    to-do tree.
    """
    with NamedTemporaryFile(mode='w+', delete=False) as file:
        writer = csv.writer(file)
        writer.writerow([REMOVE_IF_HEADER, SKIP_IF_HEADER])
        _write_template_rows(writer, todo_list_template.todos, 0)
        file.flush()
        return _upload_csv_to_google(
            '{} - {}'.format(todo_list_template.name, timezone.now()),
            file)


def import_from_spreadsheet(todo_list_template, spreadsheet_url, request):
    """Imports a Google Sheet at `spreadsheet_url` to
    `todo_list_template.todos`.

    For debugging/provenance purposes, creates a
    TodoListTemplateImportRecord with the user, todo_list_template,
    and spreadsheet URL behind the import.
    """
    try:
        reader = get_google_spreadsheet_as_csv(
            spreadsheet_url, reader=csv.reader)
    except ValueError as e:
        raise TodoListTemplateValidationError(e)
    header = next(reader)
    if header[:2] != [REMOVE_IF_HEADER, SKIP_IF_HEADER]:
        raise TodoListTemplateValidationError(
            'Unexpected header: {}'.format(header))
    # The `i`'th entry in parent_items is current list of child to-dos
    # of the `i`-depth parent. We use this to determine which parent
    # to add a child to when we read to-do in a row with lower
    # indentation than its parent.
    parent_items = []
    todos = None
    for rowindex, row in enumerate(reader):
        item = {
            'id': rowindex,
            'remove_if': json.loads(row[0] or '[]'),
            'skip_if': json.loads(row[1] or '[]'),
            'items': []
        }
        nonempty_columns = [(columnindex, text)
                            for columnindex, text in enumerate(row[2:])
                            if text]
        if len(nonempty_columns) == 0:
            continue
        elif len(nonempty_columns) > 1:
            raise TodoListTemplateValidationError(
                'More than one text entry in row {}: {}'.format(
                    rowindex, row))

        # `nonempty_index` is the depth of the column that contains
        # text in this row. We use that depth to determine which to-do
        # this is a child of.
        nonempty_index = nonempty_columns[0][0]
        item['description'] = nonempty_columns[0][1]

        if todos is None:
            # If todos has not yet been defined, `item` is the root.
            todos = item
        elif nonempty_index <= len(parent_items):
            # If the depth of the row is at most one larger than its
            # parent's depth, we can look at `parent_items` to
            # determine its parent. We truncate any parent_items entry
            # that's deeper than this item, as it is no longer
            # possible for deeper items to be parents of subsequent rows.
            parent_items = parent_items[:nonempty_index]
            # Insert this item into its parent's list of children.
            # Because child ordering is reversed in our
            # serialization format, insert it at the beginning of the
            # list of children.
            parent_items[-1].insert(0, item)
        else:
            # You can't be deeper than one column past the previous
            # row's depth.
            raise TodoListTemplateValidationError(
                'Row {} has skipped some columns in depth: {}'.format(
                    rowindex, row))
        # Add this item's child list to the end of the list of
        # parents, in case it ends up with children in subsequent rows.
        parent_items.append(item['items'])

    todo_list_template.todos = todos
    with transaction.atomic():
        todo_list_template.save()
        TodoListTemplateImportRecord.objects.create(
            import_url=spreadsheet_url,
            todo_list_template=todo_list_template,
            importer=request.user)



# -*- coding: utf-8 -*-
# Manually written
from __future__ import unicode_literals

import logging

from django.db import migrations
from django.db import models

logger = logging.getLogger(__name__)


def fill_in_worker_on_time_entries_or_delete(apps, schema_editor):
    TimeEntry = apps.get_model('orchestra', 'TimeEntry')

    for time_entry in TimeEntry.objects.all():
        if (time_entry.assignment is None or
            time_entry.assignment.worker is None):
            # Mark time entry as deleted.
            print('Deleting Time Entry {} ({} hours worked) because'.format(
                time_entry.id, time_entry.time_worked),
                  'TaskAssignment {} is None or has no worker'.format(
                      time_entry.assignment))
            time_entry.is_deleted = True
            time_entry.save()
        else:
            time_entry.worker = time_entry.assignment.worker
            time_entry.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0041_alter_payrate_fields'),
    ]

    operations = [
        migrations.RunPython(fill_in_worker_on_time_entries_or_delete),  # manually-reviewed
    ]



import json

from rest_framework import filters


class QueryParamsFilterBackend(filters.BaseFilterBackend):
    """
    Takes a json-encoded data as a queryparam from a URL,
    serializes into a python data structure
    and passes it as a queryset arguments.
    Format: ?q='{"key": ["val1", "val2"], "key2": null}', where `q` is
    filter_prefix which can be set in the view:
    filter_prefix = 'custom_key'
    Note: this doesn't support nested__fields. To support them,
    add `filterset_fields` iterable to the view.
    Note: nested__field lookup in JSONField is not supported even if added
    to the `filterset_fields`.
    This issue can be fixed when we migrate to Django 3.1
    and convert additional_data from django-jsonfields to the native one.
    """
    def _get_query_params_prefix(self, view):
        """
        By default, filter params are assigned to `q` parameter.
        """
        if hasattr(view, 'filter_prefix'):
            return view.filter_prefix
        return 'q'

    def _get_filter_kwargs(self, view, params):
        serializer = view.get_serializer(data=params)
        serializer.is_valid()
        return serializer.data

    def _get_filterset_fields_kwargs(self, view, params, qs_kwargs):
        filterset_kwargs = {}
        if hasattr(view, 'filterset_fields'):
            for field in view.filterset_fields:
                if field in params and field not in qs_kwargs:
                    filterset_kwargs[field] = params[field]
        return filterset_kwargs

    def _get_params(self, request, view):
        """
        Here we get JSON encoded params and encode them into Python objects.
        """
        params = request.query_params.copy().dict()
        json_params = params.pop(self._get_query_params_prefix(view), '{}')
        converted = json.loads(json_params)
        params.update(converted)
        return params

    def _get_kwargs(self, view, params):
        qs_kwargs = self._get_filter_kwargs(view, params)
        filterset_kwargs = self._get_filterset_fields_kwargs(
            view, params, qs_kwargs)
        qs_kwargs.update(filterset_kwargs)
        return qs_kwargs

    def filter_queryset(self, request, queryset, view):
        params = self._get_params(request, view)
        kwargs = self._get_kwargs(view, params)
        return queryset.filter(**kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import jsonfield.fields
from django.db import migrations
from django.db import models
from django.utils import timezone


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0016_auto_20150819_1900'),
    ]

    operations = [
        migrations.CreateModel(
            name='Workflow',
            fields=[
                ('id', models.AutoField(serialize=False,
                                        primary_key=True, verbose_name='ID', auto_created=True)),
                ('slug', models.CharField(max_length=200, unique=True)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('code_directory', models.CharField(max_length=255, unique=True)),
                ('sample_data_load_function', jsonfield.fields.JSONField(default={})),
            ],
        ),
        migrations.CreateModel(
            name='Step',
            fields=[
                ('id', models.AutoField(serialize=False,
                                        primary_key=True, verbose_name='ID', auto_created=True)),
                ('slug', models.CharField(max_length=200)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('is_human', models.BooleanField()),
                ('execution_function', jsonfield.fields.JSONField()),
                ('assignment_policy', jsonfield.fields.JSONField()),
                ('review_policy', jsonfield.fields.JSONField()),
                ('user_interface', jsonfield.fields.JSONField()),
                ('creation_depends_on', models.ManyToManyField(
                    to='orchestra.Step', related_name='creation_dependents', blank=True)),
            ],
        ),
        migrations.CreateModel(
            name='WorkflowVersion',
            fields=[
                ('id', models.AutoField(serialize=False,
                                        primary_key=True, verbose_name='ID', auto_created=True)),
                ('slug', models.CharField(max_length=200)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField()),
                ('workflow', models.ForeignKey(on_delete=models.CASCADE, null=True,
                                               to='orchestra.Workflow', related_name='versions')),
            ],
        ),
        migrations.AlterField(
            model_name='certification',
            name='slug',
            field=models.CharField(max_length=200),
        ),
        migrations.AlterField(
            model_name='worker',
            name='start_datetime',
            field=models.DateTimeField(default=timezone.now),
        ),
        migrations.AddField(
            model_name='step',
            name='required_certifications',
            field=models.ManyToManyField(
                to='orchestra.Certification', blank=True),
        ),
        migrations.AddField(
            model_name='step',
            name='submission_depends_on',
            field=models.ManyToManyField(
                to='orchestra.Step', related_name='submission_dependents', blank=True),
        ),
        migrations.AddField(
            model_name='step',
            name='workflow_version',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    null=True, to='orchestra.WorkflowVersion', related_name='steps'),
        ),
        migrations.AddField(
            model_name='certification',
            name='workflow',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    null=True, to='orchestra.Workflow', related_name='certifications'),
        ),
        migrations.AlterField(
            model_name='certification',
            name='required_certifications',
            field=models.ManyToManyField(
                to='orchestra.Certification', blank=True, related_name='dependent_certifications'),
        ),
        migrations.AddField(
            model_name='project',
            name='workflow_version',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    null=True, to='orchestra.WorkflowVersion', related_name='projects'),
        ),
        migrations.AddField(
            model_name='task',
            name='step',
            field=models.ForeignKey(on_delete=models.CASCADE,
                                    null=True, to='orchestra.Step', related_name='tasks'),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='workflowversion',
            unique_together=set([('workflow', 'slug')]),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='step',
            unique_together=set([('workflow_version', 'slug')]),
        ),
        migrations.AlterUniqueTogether(  # manually-reviewed
            name='certification',
            unique_together=set([('workflow', 'slug')]),
        ),
    ]



import logging
import random
import string

from django.conf import settings
from django.utils.text import slugify

from requests.exceptions import HTTPError

from slacker import Error as SlackError
from slacker import BaseAPI
from slacker import Slacker

from orchestra.communication.errors import SlackFormatError
from orchestra.utils.decorators import run_if

logger = logging.getLogger(__name__)

# Types of responses we can send to slack
VALID_RESPONSE_TYPES = {'ephemeral', 'in_channel'}
_request = BaseAPI._request


def _silent_request(*args, **kwargs):
    """
    Attempt to make Slack API request and ignore if an exception is thrown.

    TODO(jrbotros): this silences all errors, but we likely will want to be
    able to surface errors in some cases in the future
    """
    try:
        return _request(*args, **kwargs)
    except SlackError:
        logger.exception('Slack API Error')
    except HTTPError as e:
        status_code = e.response.status_code
        # If we're being rate-limited, log the exception but don't fail.
        if status_code == 429:
            logger.exception('Slack API rate limit')
        else:
            raise


BaseAPI._request = _silent_request


class OrchestraSlackService(object):
    """
    Wrapper slack service to allow easy swapping and mocking out of API.
    """

    def __init__(self, api_key=None):
        if not api_key:
            api_key = settings.SLACK_EXPERTS_API_KEY
        self._service = Slacker(api_key)
        for attr_name in ('chat', 'groups', 'users'):
            setattr(self, attr_name, getattr(self._service, attr_name))


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def get_slack_user_id(slack_username):
    slack = OrchestraSlackService()
    slack_user_id = slack.users.get_user_id(slack_username)
    return slack_user_id


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def add_worker_to_project_team(worker, project):
    slack = OrchestraSlackService()
    try:
        user_id = slack.users.get_user_id(worker.slack_username)
        response = slack.groups.invite(project.slack_group_id, user_id)
        if not response.body.get('already_in_group'):
            welcome_message = (
                '<@{}|{}> has been added to the team. '
                'Welcome aboard!').format(user_id, worker.slack_username)
            slack.chat.post_message(project.slack_group_id, welcome_message)
    except SlackError:
        logger.exception('Slack API Error')
        # TODO(jrbotros): for now, using slack on a per-worker basis is
        # optional; we'll want to rethink this in the future


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def create_project_slack_group(project):
    """
    Create slack channel for project team communication
    """
    slack = OrchestraSlackService()
    response = slack.groups.create(_project_slack_group_name(project))
    project.slack_group_id = response.body['group']['id']
    slack.groups.set_topic(project.slack_group_id, project.short_description)
    slack.groups.set_purpose(project.slack_group_id,
                             'Discussing work on `{}`'.format(
                                 project.short_description))
    project.save()

    # Message out project folder id.
    if project.project_data.get('project_folder_id'):
        message = (
            'Project folder: '
            'https://drive.google.com/drive/folders/{}'
        ).format(project.project_data['project_folder_id'])
        slack.chat.post_message(project.slack_group_id, message)
    return project.slack_group_id


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def archive_project_slack_group(project):
    """
    Archive a slack channel of a project
    """
    slack = OrchestraSlackService()
    try:
        response = slack.groups.archive(project.slack_group_id)
        if response:
            is_archived = response.body.get('ok')
            if not is_archived:
                logger.error('Archive project error: %s',
                             response.body.get('error'))
    except SlackError:
        logger.exception('Slack API Error')


@run_if('ORCHESTRA_SLACK_EXPERTS_ENABLED')
def unarchive_project_slack_group(project):
    """
    Unarchive a slack channel of a project
    """
    slack = OrchestraSlackService()
    try:
        group_info = slack.groups.info(project.slack_group_id)
        is_archived = group_info.body.get('group', {}).get('is_archived')
        if is_archived:
            response = slack.groups.unarchive(project.slack_group_id)
            if response:
                is_unarchived = response.body.get('ok')
                if not is_unarchived:
                    logger.error('Unarchive project error: %s',
                                 response.body.get('error'))
    except SlackError:
        logger.exception('Slack API Error')


def _random_string():
    return ''.join(
        random.choice(string.ascii_lowercase + string.digits)
        for idx in range(4))


def _project_slack_group_name(project):
    """
    Return a unique and readable identifier for project slack groups.

    Slack group names are capped at 21 characters in length.
    """
    name = None
    # The human-readable portion of the name (16 characters) involves
    # slugifying the project short description.
    descriptor = slugify(project.short_description)[:16].strip('-')
    slack = OrchestraSlackService()
    groups = {group['name'] for group in slack.groups.list().body['groups']}
    while True:
        # Add 4 characters of randomness (~1.68 million permutations).
        name = '{}-{}'.format(descriptor, _random_string())
        if name not in groups:
            break
    return name


def format_slack_message(text,
                         attachments=None,
                         response_type='ephemeral'):
    """
    Args:
        text (str):
            Plain text message to send
        attachments (dict):
           See https://api.slack.com/docs/attachments
        response_type (string):
            Should be `in_channel` or `ephemeral`. See
            https://api.slack.com/slash-commands
    Returns:
        formatted_message (dict):
            A formatted message to send via the slack client
    """
    if response_type not in VALID_RESPONSE_TYPES:
        raise SlackFormatError(
            'Response type {} is invalid'.format(response_type))
    return {
        'response_type': response_type,
        'text': text,
        'attachments': attachments,
    }



from django.conf.urls import url
from rest_framework import routers

from orchestra.todos.views import TodoQADetail
from orchestra.todos.views import TodoQAList
from orchestra.todos.views import TodoViewset
from orchestra.todos.views import TodoListTemplateDetail
from orchestra.todos.views import TodoListTemplateList
import orchestra.todos.views as views

app_name = 'todos'

urlpatterns = [
    url(r'^todo_qa/$',
        TodoQAList.as_view(), name='todo_qas'),
    url(r'^todo_qa/(?P<pk>[0-9]+)/$',
        TodoQADetail.as_view(), name='todo_qa'),
    url(r'^todolist_templates/$',
        TodoListTemplateList.as_view(), name='todolist_templates'),
    url(r'^todolist_template/(?P<pk>[0-9]+)/$',
        TodoListTemplateDetail.as_view(), name='todolist_template'),
    url(r'^update_todos_from_todolist_template/$',
        views.update_todos_from_todolist_template,
        name='update_todos_from_todolist_template'),
    url(r'^worker_task_recent_todo_qas/$',
        views.worker_task_recent_todo_qas,
        name='worker_task_recent_todo_qas'),
    url(r'^todolist_template/(?P<pk>[0-9]+)/import$',
        views.ImportTodoListTemplateFromSpreadsheet.as_view(),
        name='import_todo_list_template_from_spreadsheet'),
]

router = routers.SimpleRouter()
router.register(
    r'todo', TodoViewset, basename='todos'
)

urlpatterns += router.urls



# Generated by Django 2.2.9 on 2020-03-30 19:21

from django.db import migrations, models

def migrate_request_status(apps, schema_editor):
    StaffBotRequest = apps.get_model('orchestra', 'StaffBotRequest')
    StaffingResponse = apps.get_model('orchestra', 'StaffingResponse')
    StaffingRequestInquiry = apps.get_model('orchestra', 'StaffingRequestInquiry')

    DONE_SENDING_INQUIRIES = 1
    CLOSED = 2

    requests = (StaffBotRequest.objects.all()
                .exclude(status=CLOSED))
    for request in requests:
        responses = StaffingResponse.objects.filter(
                        request_inquiry__request=request)
        has_winner = responses.filter(is_winner=True).exists()
        if has_winner:
            request.status = CLOSED
        elif request.status == DONE_SENDING_INQUIRIES:
            inquiries = (
                StaffingRequestInquiry.objects.filter(request=request)
            ).distinct()
            num_inquired_workers = len(
                set(inquiries.values_list(
                    'communication_preference__worker__id', flat=True)
                    )
            )
            responded_inquiries = inquiries.filter(
                responses__isnull=False).distinct()
            num_responded_workers = len(
                set(responded_inquiries.values_list(
                    'communication_preference__worker__id', flat=True)
                    )
            )

            if num_responded_workers >= num_inquired_workers:
                request.status = CLOSED
    StaffBotRequest.objects.bulk_update(list(requests), ['status'])


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0082_task_tags'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staffbotrequest',
            name='status',
            field=models.IntegerField(choices=[(0, 'sending inquiries'), (1, 'done sending inquiries'), (2, 'closed')], default=0),  # append-only
        ),
        migrations.RunPython(migrate_request_status),  # manually-reviewed
    ]



from unittest.mock import patch

from django.core.management import call_command
from django.core.management.base import CommandError

from orchestra.tests.helpers import OrchestraTestCase


class MigrateCertificationsTestCase(OrchestraTestCase):
    patch_path = ('orchestra.management.commands.'
                  'migrate_certifications.migrate_certifications')

    @patch(patch_path)
    def test_options(self, mock_migrate):
        # Test no options
        with self.assertRaises(CommandError):
            call_command('migrate_certifications')
            mock_migrate.assert_not_called()

        # Test
        call_command('migrate_certifications',
                     'test_source_workflow_slug',
                     'ntest_destination_workflow_slug',
                     certifications=['test_cert_1', 'test_cert_2'])
        mock_migrate.called_once_with(
            'test_source_workflow_slug',
            'test_destination_workflow_slug',
            ['test_cert_1', 'test_cert_2']
        )



from rest_framework.test import APIRequestFactory
from rest_framework.viewsets import ModelViewSet

from orchestra.models import Todo
from orchestra.models import TodoQA
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.orchestra_api import build_url_params
from orchestra.todos.filters import QueryParamsFilterBackend
from orchestra.todos.serializers import TodoQASerializer
from orchestra.tests.helpers import OrchestraTransactionTestCase
from orchestra.todos.views import GenericTodoViewset


factory = APIRequestFactory()


class DummyView(ModelViewSet):
    model = TodoQA
    serializer_class = TodoQASerializer
    queryset = TodoQA.objects.all()


class QueryParamsFilterBackendTests(OrchestraTransactionTestCase):
    def _get_qs_kwargs(self, view_class, url_params):
        url = '/some-path/{}'.format(url_params)
        view = view_class(action_map={'get': 'list'})
        request = factory.get(url)
        request = view.initialize_request(request)
        view.request = request
        view.format_kwarg = {}
        backend = QueryParamsFilterBackend()
        params = backend._get_params(request, view)
        kwargs = backend._get_kwargs(view, params)
        return kwargs

    def test_random_key_values(self):
        url_params = build_url_params(
            123, 'slug', **{'some_random_key': 'some_value'})
        kwargs = self._get_qs_kwargs(DummyView, url_params)
        self.assertEqual(kwargs, {})

    def test_only_serializer_fields_are_passed(self):
        """
        In GenericTodoViewset we have the following filterset_fields:
        filterset_fields = ('project__id', 'step__slug',)
        A field which is not specified in it or in the view's
        serializer fields cannot be passed as a filter args
        """
        title = 'Some title'
        project_id = 123
        url_params = build_url_params(
            project_id,
            None,
            **{
                'title': title,
                'nonexistent_field': True})
        kwargs = self._get_qs_kwargs(
            GenericTodoViewset, url_params)
        self.assertEqual(
            kwargs, {'project__id': str(project_id), 'title': title})

    def test_dangerous_sql(self):
        project = ProjectFactory()
        TodoFactory(project=project)
        dangerous_sql = (
            "if ((select user) = 'sa'"
            " OR (select user) = 'dbo') select 1 else select 1/0"
        )
        url_params = build_url_params(
            project.id,
            None,
            **{'additional_data__sql': dangerous_sql}
        )
        kwargs = self._get_qs_kwargs(
            GenericTodoViewset, url_params)
        # kwargs doesn't contain additional_data__sql field
        self.assertEqual(kwargs, {'project__id': str(project.id)})

        url_params = build_url_params(
            project.id,
            None,
            **{'title': dangerous_sql})
        kwargs = self._get_qs_kwargs(
            GenericTodoViewset, url_params)
        todos = Todo.objects.filter(**kwargs)
        self.assertEqual(
            kwargs,
            {'project__id': str(project.id), 'title': dangerous_sql})
        self.assertTrue(todos.count, 0)



import logging

from orchestra.google_apps.convenience import create_document_from_template
from orchestra.google_apps.convenience import create_folder_with_permissions
from orchestra.google_apps.permissions import write_with_link_permission

logger = logging.getLogger(__name__)


def create_documents(project_data, prerequisites):
    """Create documents and folders needed for the journalism workflow.

    The following will be created:
    * an 'Article Draft' document where a reporter can draft text.
    * a 'Raw Photos' folder where a photographer can upload images.

    Documents are created in the project root folder.
    """
    task_data = {}
    folder_id = project_data['project_folder_id']

    # Create an Article Draft document.
    article_draft_template = project_data['article_draft_template']
    task_data['articleURL'] = create_document_from_template(
        article_draft_template,
        'Article Draft',
        parent_ids=[folder_id],
        permissions=[write_with_link_permission],
    )['alternateLink']

    # Create a Raw Photos folder.
    task_data['raw_photo_folder'] = create_folder_with_permissions(
        folder_id,
        'Raw Photos',
        permissions=[write_with_link_permission],
    )['id']

    return task_data



import json
from unittest.mock import patch

from django.test import TestCase
from rest_framework.test import APIClient

from orchestra.models import Todo
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import WorkflowVersionFactory
from orchestra.project_api.auth import SignedUser
from orchestra.orchestra_api import create_todos
from orchestra.orchestra_api import get_todos
from orchestra.orchestra_api import update_todos
from orchestra.orchestra_api import delete_todos
from orchestra.orchestra_api import OrchestraError


class TodoAPITests(TestCase):
    def setUp(self):
        super().setUp()
        self.request_client = APIClient(enforce_csrf_checks=True)
        self.request_client.force_authenticate(user=SignedUser())
        self.workflow_version = WorkflowVersionFactory()
        self.step = StepFactory(
            slug='step-slug',
            workflow_version=self.workflow_version)
        self.project = ProjectFactory(
            workflow_version=self.workflow_version)

    @patch('orchestra.orchestra_api.requests')
    def test_create_todos(self, mock_request):
        # This converts `requests.post` into DRF's `APIClient.post`
        # To make it testable
        def post(url, *args, **kwargs):
            kw = kwargs.get('data', '')
            data = json.loads(kw)
            return_value = self.request_client.post(url, data, format='json')
            return_value.text = json.dumps(return_value.data)
            return return_value

        mock_request.post = post
        data = [
            {
                'title': 'Testing title {}'.format(x),
                'project': self.project.id,
                'step': self.step.slug
            } for x in range(10)
        ]
        result = create_todos(data)
        self.assertEqual(len(result), 10)
        for r in result:
            self.assertTrue(r['title'].startswith('Testing title'))

    @patch('orchestra.orchestra_api.requests')
    def test_get_todos(self, mock_request):
        # This converts `requests.get` into DRF's `APIClient.get`
        # To make it testable
        def get(url, *args, **kwargs):
            return_value = self.request_client.get(url, format='json')
            return_value.text = json.dumps(return_value.data)
            return return_value

        mock_request.get = get

        step_1 = StepFactory(slug='step_1')
        step_2 = StepFactory(slug='step_2')
        project2 = ProjectFactory()
        # 2 todos have the same project
        todo1 = TodoFactory(step=step_1, project=self.project)
        todo2 = TodoFactory(step=step_1, project=self.project)
        # 1 todo has another project
        todo3 = TodoFactory(step=step_2, project=project2)

        todo4 = TodoFactory(step=step_1, project=project2)

        # Get todo1 and todo2
        res = get_todos(self.project.id, step_1.slug)
        self.assertEqual(len(res), 2)
        expected_ids = [todo1.id, todo2.id]
        for r in res:
            self.assertIn(r['id'], expected_ids)

        # Get todo3
        res = get_todos(project2.id, step_2.slug)
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0]['id'], todo3.id)

        # Filter by project_id
        res = get_todos(project_id=project2.id)
        self.assertEqual(len(res), 2)
        expected_ids = [todo3.id, todo4.id]
        for r in res:
            self.assertIn(r['id'], expected_ids)

        # Test project_id is required
        msg = 'project_id is required'
        with self.assertRaisesMessage(OrchestraError, msg):
            get_todos(None)

    @patch('orchestra.orchestra_api.requests')
    def test_update_todos(self, mock_request):
        # This converts `requests.patch` into DRF's `APIClient.patch`
        # To make it testable
        def patch_func(url, *args, **kwargs):
            kw = kwargs.get('data', '')
            data = json.loads(kw)
            return_value = self.request_client.put(
                url, data, format='json')
            return_value.text = json.dumps(return_value.data)
            return return_value

        mock_request.patch = patch_func

        todo1 = TodoFactory(step=self.step, project=self.project)
        todo2 = TodoFactory(step=self.step, project=self.project)
        todo3 = TodoFactory(step=self.step, project=self.project)
        todo_should_not_be_updated = TodoFactory(
            project=self.project, step=self.step, title='Not updated')
        # Change titles
        todos_with_updated_titles = [{
            'id': x.id,
            'title': 'Updated title {}'.format(x.id),
            'step': x.step.slug,
            'project': x.project.id
        } for x in [todo1, todo3, todo2]]
        result = update_todos(todos_with_updated_titles)
        self.assertEqual(len(result), 3)
        updated_todos = Todo.objects.filter(
            id__in=[todo1.id, todo2.id, todo3.id])
        for todo in updated_todos:
            self.assertEqual(todo.title, 'Updated title {}'.format(todo.id))
        self.assertEqual(todo_should_not_be_updated.title, 'Not updated')

    def _change_attr(self, item, attr, value):
        item[attr] = value
        return item

    @patch('orchestra.orchestra_api.requests')
    def test_delete_todos(self, mock_request):
        # This converts `requests.delete` into DRF's `APIClient.delete`
        # To make it testable
        def delete(url, *args, **kwargs):
            kw = kwargs.get('data', '')
            data = json.loads(kw)
            return_value = self.request_client.delete(
                url, data, format='json')
            return_value.text = json.dumps(return_value.data)
            return return_value

        mock_request.delete = delete

        todo1 = TodoFactory(step=self.step, project=self.project)
        todo2 = TodoFactory(step=self.step, project=self.project)
        todo3 = TodoFactory(step=self.step, project=self.project)
        todo4 = TodoFactory(step=self.step, project=self.project)

        res = delete_todos([todo1.id, todo2.id, todo3.id])
        self.assertEqual(res[0], 3)
        left_todos = Todo.objects.all()
        self.assertEqual(left_todos.count(), 1)
        self.assertEqual(left_todos[0].id, todo4.id)



# -*- coding: utf-8 -*-
# Manually written
from __future__ import unicode_literals

from django.db import migrations


def copy_start_datetime_to_created_at(apps, schema_editor):
    Iteration = apps.get_model('orchestra', 'Iteration')
    TaskAssignment = apps.get_model('orchestra', 'TaskAssignment')

    for iteration in Iteration.objects.all():
        iteration.created_at = iteration.start_datetime
        iteration.save()

    for assignment in TaskAssignment.objects.all():
        assignment.created_at = assignment.start_datetime
        assignment.save()


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0038_add_fields_to_taskassignment'),
    ]

    operations = [
        migrations.RunPython(copy_start_datetime_to_created_at),  # manually-reviewed
    ]



from django.conf.urls import include
from django.conf.urls import url

from orchestra.views import index
from orchestra.views import newindex
from orchestra.views import status

app_name = 'orchestra'

urlpatterns = [
    url(r'^api/',
        include('orchestra.api_urls', namespace='orchestra')),
    url(r'^communication/',
        include('orchestra.communication.urls',
                namespace='communication')),
    url(r'^newapp/?', newindex, name='newindex'),
    url(r'^app/?', index, name='index'),
    url(r'', include('orchestra.accounts.urls')),
    url(r'^bots/', include('orchestra.bots.urls', namespace='bots')),
    url(r'^todos/', include('orchestra.todos.urls', namespace='todos')),

    # Health check status
    url(r'^status/', status, name='status'),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-05-05 13:49
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ('orchestra', '0059_step_description_function'),
    ]

    operations = [
        migrations.RenameModel(
            old_name='StaffingRequest',
            new_name='StaffingRequestInquiry',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.5 on 2016-04-06 21:57
from __future__ import unicode_literals

import logging
from operator import itemgetter

from dateutil.parser import parse
from django.db import migrations
from django.db import models

# NOTE(jrbotros): change import or copy into migration if code moves/changes
from orchestra.tests.helpers.iterations import verify_iterations

logger = logging.getLogger(__name__)


class SnapshotType:
    SUBMIT = 0
    ACCEPT = 1
    REJECT = 2


class IterationStatus:
    PROCESSING = 0
    REQUESTED_REVIEW = 1
    PROVIDED_REVIEW = 2


class AssignmentStatus:
    PROCESSING = 0
    SUBMITTED = 1
    FAILED = 2


# Map snapshot types onto iteration statuses
snapshot_type_to_iteration_status = {
    SnapshotType.SUBMIT: IterationStatus.REQUESTED_REVIEW,
    SnapshotType.ACCEPT: IterationStatus.REQUESTED_REVIEW,
    SnapshotType.REJECT: IterationStatus.PROVIDED_REVIEW,
}


def get_ordered_snapshots(task):
    task_snapshots = []
    for assignment in task.assignments.all():
        assignment_snapshots = assignment.snapshots.get('snapshots')
        if assignment_snapshots is None:
            continue
        for assignment_snap_index, snapshot in enumerate(assignment_snapshots):
            snapshot['assignment'] = assignment
            # Keep track of the snapshot's index relative to the assignment
            snapshot['assignment_snap_index'] = assignment_snap_index
            task_snapshots.append(snapshot)
    return sorted(task_snapshots, key=itemgetter('datetime'))


def snapshots_to_iterations(apps, schema_editor):
    Iteration = apps.get_model('orchestra', 'Iteration')
    Task = apps.get_model('orchestra', 'Task')
    TaskAssignment = apps.get_model('orchestra', 'TaskAssignment')

    for task in Task.objects.all():
        task_snapshots = get_ordered_snapshots(task)
        for task_snap_index, snapshot in enumerate(task_snapshots):
            assignment = snapshot['assignment']
            iteration = Iteration.objects.create(
                assignment=assignment,
                end_datetime=parse(snapshot['datetime']),
                submitted_data=snapshot['data'],
                status=snapshot_type_to_iteration_status[snapshot['type']])
            if snapshot['assignment_snap_index'] == 0:
                # Snapshot is the first for its assignment, so its start
                # datetime will be the same as its assignment
                # NOTE: This should cover the case where task_snap_index == 0
                iteration.start_datetime = assignment.start_datetime
            else:
                iteration.start_datetime = (
                    task_snapshots[task_snap_index - 1]['datetime'])
            iteration.save()

        processing_assignments = task.assignments.filter(
            status=AssignmentStatus.PROCESSING)
        if processing_assignments.exists():
            if processing_assignments.count() > 1:
                logger.error(
                    'Task {} has too many processing assignments'.format(
                        task.id))
            else:
                # If task has a processing assignment, create a final
                # processing iteration
                processing_assignment = processing_assignments.first()
                if (not task_snapshots
                        or not processing_assignment.iterations.exists()):
                    final_start_datetime = processing_assignment.start_datetime
                else:
                    final_start_datetime = task_snapshots[-1]['datetime']
                iteration = Iteration.objects.create(
                    assignment=processing_assignment,
                    start_datetime=final_start_datetime,
                    status=IterationStatus.PROCESSING)

        try:
            verify_iterations(task.id)
        except AssertionError:
            logger.error('Iterations invalid for task {}.'.format(task.id))


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0027_create_time_entries_for_snapshots'),
    ]

    operations = [
        migrations.AlterField(
            model_name='iteration',
            name='status',
            field=models.IntegerField(choices=[(
                0, 'Processing'), (1, 'Requested Review'), (2, 'Provided Review')], default=0),
        ),
        migrations.RunPython(snapshots_to_iterations)  # manually-reviewed
    ]



import logging
from urllib.parse import urljoin

from django.conf import settings
from django import urls
from django.utils import timezone

from orchestra.communication.slack import OrchestraSlackService
from orchestra.models import Project
from orchestra.project_api.api import get_project_information

logger = logging.getLogger(__name__)


def project_management_information(project_id):
    project = Project.objects.get(id=project_id)
    project_information = get_project_information([project.id])
    project_information[project.id]['project']['status'] = dict(
        Project.STATUS_CHOICES).get(project.status, None)
    project_information[project.id]['project']['admin_url'] = urljoin(
        settings.ORCHESTRA_URL,
        urls.reverse(
            'admin:orchestra_project_change',
            args=(project_id,)))

    for slug, task in project_information[project.id]['tasks'].items():
        task['admin_url'] = urljoin(
            settings.ORCHESTRA_URL,
            urls.reverse(
                'admin:orchestra_task_change',
                args=(task['id'],)))

        for assignment in task['assignments']:
            assignment['admin_url'] = urljoin(
                settings.ORCHESTRA_URL,
                urls.reverse(
                    'admin:orchestra_taskassignment_change',
                    args=(assignment['id'],)))

            for iteration in assignment['iterations']:
                iteration['admin_url'] = urljoin(
                    settings.ORCHESTRA_URL,
                    urls.reverse(
                        'admin:orchestra_iteration_change',
                        args=(iteration['id'],)))
                if not iteration['end_datetime']:
                    iteration['end_datetime'] = timezone.now().isoformat()

    return project_information


def edit_slack_membership(project_id, username, action):
    slack = OrchestraSlackService()
    slack_user_id = slack.users.get_user_id(username)
    slack_group_id = Project.objects.get(id=project_id).slack_group_id
    if action == 'add':
        slack.groups.invite(slack_group_id, slack_user_id)
    elif action == 'remove':
        slack.groups.kick(slack_group_id, slack_user_id)
    else:
        raise Exception('Action not found.')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0006_auto_20150501_1809'),
    ]

    operations = [
        migrations.AddField(
            model_name='taskassignment',
            name='assignment_counter',
            field=models.IntegerField(default=0),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='certification',
            name='required_certifications',
            field=models.ManyToManyField(
                to='orchestra.Certification', related_name='required_certifications_rel_+', blank=True, null=True),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='project',
            name='process_slug',
            field=models.CharField(
                max_length=200, choices=[('doctors', 'Doctors Process')]),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='task',
            name='step_slug',
            field=models.CharField(max_length=200, choices=[
                                   ('content_extraction', ' Content Extraction'), ('copy_pass', 'Copy Pass')]),
            preserve_default=True,
        ),
    ]



from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from jsonschema import Draft4Validator
from jsonschema.exceptions import ValidationError as SchemaValidationError
from django.utils.translation import gettext_lazy
from orchestra.utils.json_schema import DefaultValidatingDraft4Validator


class JSONField(JSONField):
    default_error_messages = {
        'schema_invalid': gettext_lazy(
            'Value failed schema validation: %(err)s'),
    }

    def __init__(self, schema=None, add_defaults=False, *args, **kwargs):
        self.schema = schema
        self.add_defaults = add_defaults
        if self.schema and not isinstance(self.schema, dict):
            raise ValueError('Schema must be a dict object.')
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # Only include kwarg if it's not the default
        if self.schema:
            kwargs['schema'] = self.schema
        if self.add_defaults:
            kwargs['add_defaults'] = self.add_defaults
        return name, path, args, kwargs

    def get_prep_value(self, value):
        self._validate_with_schema(value)
        return super().get_prep_value(value)

    def _validate_with_schema(self, value):
        if value and self.schema:
            try:
                self._get_validator_class()(self.schema).validate(value)
            except SchemaValidationError as err:
                raise ValidationError(
                    self.error_messages['schema_invalid'],
                    code='schema_invalid',
                    params={'err': err},
                )

    def _get_validator_class(self):
        if self.add_defaults:
            return DefaultValidatingDraft4Validator
        else:
            return Draft4Validator



import logging

from django.db import transaction

from orchestra.models import StaffBotRequest
from orchestra.models import StaffingResponse

logger = logging.getLogger(__name__)


@transaction.atomic
def mark_worker_as_winner(worker, task, required_role_counter,
                          staffing_request_inquiry):
    staffbot_request = (
        StaffBotRequest.objects
        .filter(task=task, required_role_counter=required_role_counter)
        .exclude(status=StaffBotRequest.Status.CLOSED.value)
        .order_by('-created_at'))

    # Check whether staffbot request was sent out for this task
    if not staffbot_request.exists():
        return

    if staffing_request_inquiry:
        staffbot_request = staffing_request_inquiry.request
    else:
        staffbot_request = staffbot_request.first()

    close_open_staffbot_requests(task)

    # If staffing request inquiry provided
    if staffing_request_inquiry:
        staffing_response = staffing_request_inquiry.responses.all()
    else:
        staffing_response = StaffingResponse.objects.filter(
            request_inquiry__request=staffbot_request,
            request_inquiry__communication_preference__worker=worker)

    if staffing_response.exists():
        staffing_response = staffing_response.first()
        staffing_response.is_available = True
        staffing_response.is_winner = True
        staffing_response.save()


@transaction.atomic
def close_open_staffbot_requests(task):
    CLOSED = StaffBotRequest.Status.CLOSED.value
    requests = task.staffing_requests.all()
    for request in requests:
        request.status = CLOSED
    StaffBotRequest.objects.bulk_update(requests, ['status'])



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.utils.timezone
import jsonfield.fields
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0019_auto_20151007_1846'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taskassignment',
            name='start_datetime',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name='step',
            name='assignment_policy',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AlterField(
            model_name='step',
            name='execution_function',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AlterField(
            model_name='step',
            name='review_policy',
            field=jsonfield.fields.JSONField(default={}),
        ),
        migrations.AlterField(
            model_name='step',
            name='user_interface',
            field=jsonfield.fields.JSONField(default={}),
        ),
    ]



from django.conf import settings

from orchestra.interface_api.project_management.decorators import \
    is_project_admin


def google_analytics(request):
    """
    Provide the Google Analytics setting to any template that needs it.
    """
    return {
        'GOOGLE_ANALYTICS_KEY': settings.GOOGLE_ANALYTICS_KEY,
    }


def third_party_scripts(request):
    """
    Provide the path to common third party scripts to any template that needs
    it.
    """
    return {
        'ORCHESTRA_THIRD_PARTY_SCRIPTS_TEMPLATE':
        settings.ORCHESTRA_THIRD_PARTY_SCRIPTS_TEMPLATE
    }


def base_context(request):
    """
    Provide context variables for use across all views.
    """
    if not hasattr(request, 'user'):
        return {}
    return {
        'is_project_admin': is_project_admin(request.user)
    }



from pprint import pprint

from django.core.management.base import BaseCommand

from orchestra.orchestra_api import create_orchestra_project
from orchestra.orchestra_api import get_project_information


class Command(BaseCommand):
    help = 'Start and examine projects using the journalism workflow'

    def add_arguments(self, parser):
        parser.add_argument(
            '-p', '--project-id',
            help='Existing project id to check status of')
        primary_command_group = parser.add_mutually_exclusive_group(
            required=True)
        primary_command_group.add_argument(
            '-n', '--new', action='store_true',
            help='Start a new journalism project')
        primary_command_group.add_argument(
            '-s', '--status', action='store_true',
            help='Check the status of an in-progress project')
        primary_command_group.add_argument(
            '-f', '--final', action='store_true',
            help='Output a final summary of the project')

    def handle(self, *args, **options):
        if not self.verify_options(options):
            return

        if options['new']:
            project_id = self.create_project()
            self.project_info(project_id)

        elif options['status']:
            self.project_info(options['project_id'])

        elif options['final']:
            project_id = options['project_id']
            p_info = self.project_info(project_id, verbose=False)
            try:
                tasks = p_info[project_id]['tasks']
                assert tasks['copy_editing']['status'] == 'Complete'
                pprint(tasks['copy_editing']['latest_data'])
            except Exception:
                print("The '--final' option must be used with a completed "
                      "project")

    def create_project(self):
        project_id = create_orchestra_project(
            None,
            'journalism',
            'v1',
            'A test run of our journalism workflow',
            10,
            {
                'article_draft_template': '1F9ULJ_eoJFz1whqjK2thsC6gJup2f35IsUUpbcizcfA'  # noqa
            },
            'train',
        )
        print('Project with id {} created!'.format(project_id))
        return project_id

    def project_info(self, project_id, verbose=True):
        p_info = get_project_information([project_id])
        if verbose:
            print("Information received! Here's what we got:")
            print('')
            pprint(p_info)
        return p_info

    def verify_options(self, options):
        if options['new'] and options['project_id']:
            print("The '--new' option cannot be used in conjunction with a "
                  "project id.")
            return False

        elif options['status'] and not options['project_id']:
            print("The '--status' option must be used in conjunction with a "
                  "project id.")
            return False

        elif options['final'] and not options['project_id']:
            print("The '--final' option must be used in conjunction with a "
                  "project id.")
            return False
        return True



import json
import datetime
from unittest.mock import patch

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.urls import reverse
from django.test import override_settings
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.test import APIClient

from orchestra.google_apps.service import Service
from orchestra.models import Project
from orchestra.models import Task
from orchestra.models import Todo
from orchestra.models import TaskAssignment
from orchestra.models import TimeEntry
from orchestra.models import Worker
from orchestra.models import WorkerCertification
from orchestra.todos.serializers import BulkTodoSerializer
from orchestra.project_api.api import MalformedDependencyException
from orchestra.project_api.api import get_workflow_steps
from orchestra.project_api.api import get_project_information
from orchestra.project_api.auth import OrchestraProjectAPIAuthentication
from orchestra.project_api.auth import SignedUser
from orchestra.tests.helpers import OrchestraTestCase
from orchestra.tests.helpers import EndpointTestCase
from orchestra.tests.helpers.fixtures import setup_models
from orchestra.tests.helpers.fixtures import StepFactory
from orchestra.tests.helpers.fixtures import ProjectFactory
from orchestra.tests.helpers.fixtures import TodoFactory
from orchestra.tests.helpers.fixtures import WorkflowVersionFactory
from orchestra.tests.helpers.google_apps import mock_create_drive_service
from orchestra.utils.load_json import load_encoded_json
from orchestra.utils.task_lifecycle import get_new_task_assignment


class ProjectAPITestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.api_client = APIClient(enforce_csrf_checks=True)
        self.api_client.force_authenticate(user=SignedUser())

    def test_project_details_url(self):
        project = Project.objects.all()[0]
        response = self.api_client.post(
            reverse('orchestra:orchestra:project_details_url'),
            {'project_id': project.id},
            format='json')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)
        project_details_url = returned.get('project_details_url')
        self.assertEqual(project_details_url,
                         ('http://testserver/orchestra/app/project/%s' %
                          project.id))

        response = self.api_client.post(
            '/orchestra/api/project/project_details_url/',
            {},
            format='json')

        self.ensure_response(response,
                             {'error': 400,
                              'message': 'project_id parameter is missing'},
                             400)

    def test_project_information(self):
        project = self.projects['base_test_project']
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_ids': [project.id]},
            format='json')
        self.assertEqual(response.status_code, 200)
        returned = load_encoded_json(response.content)

        self._delete_keys(returned[str(project.id)])
        for item in returned.values():
            del item['tasks']['step1']['project']
            del (item['tasks']['step1']['assignments'][0]
                     ['iterations'][0]['assignment'])

        expected = {
            'project': {
                'task_class': 1,
                'workflow_slug': 'w1',
                'workflow_version_slug': 'test_workflow',
                'project_data': {},
                'team_messages_url': None,
                'priority': 0,
                'status': 0
            },
            'tasks': {
                'step1': {
                    'assignments': [{
                        'status': 'Submitted',
                        'in_progress_task_data': {'test_key': 'test_value'},
                        'worker': {
                            'username': self.workers[0].user.username,
                            'first_name': self.workers[0].user.first_name,
                            'last_name': self.workers[0].user.last_name,
                            'slack_username':
                                self.workers[0].slack_username,
                            'slack_user_id':
                                self.workers[0].slack_user_id
                        },
                        'iterations': [{
                            'status': 'Requested Review',
                            'submitted_data': {'test_key': 'test_value'},
                        }],
                        'recorded_work_time': 30*60,
                    }],
                    'latest_data': {
                        'test_key': 'test_value'
                    },
                    'status': 'Pending Review',
                    'step_slug': 'step1',
                }
            },
            'steps': [
                {'slug': 'step1',
                 'description': 'The longer description of the first step',
                 'is_human': True,
                 'name': 'The first step'},
                {'slug': 'step2',
                 'description': 'The longer description of the second step',
                 'is_human': True,
                 'name': 'The second step'},
                {'slug': 'step3',
                 'description': 'The longer description of the third step',
                 'is_human': True,
                 'name': 'The third step'}
            ]
        }

        self.assertEqual(returned[str(project.id)], expected)

        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_ids': [-1]},
            format='json')
        self.assertEqual(load_encoded_json(response.content), {})

        # Getting project info without a project_ids should fail.
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': [project.id]},  # Typo.
            format='json')
        msg = 'project_ids is required'
        self.ensure_response(response,
                             {'error': 400, 'message': msg}, 400)

        # Retrieve the third project, which has no task assignments.
        project_id = self.projects['no_task_assignments'].id
        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_ids': [project_id]},
            format='json')
        returned = load_encoded_json(response.content)
        for key in ('id', 'project', 'start_datetime'):
            del returned[str(project_id)]['tasks']['step1'][key]
        self.assertEqual(response.status_code, 200)
        self.assertEqual(returned[str(project_id)]['tasks'], {
            'step1': {
                'assignments': [],
                'latest_data': None,
                'status': 'Awaiting Processing',
                'step_slug': 'step1'
            }
        })

    def test_project_assignment_recorded_time(self):
        project = self.projects['base_test_project']
        worker = self.workers[0]

        task = self.tasks['review_task']
        assignment = TaskAssignment.objects.filter(
            worker=worker, task=task).first()
        other_assignment = get_new_task_assignment(
            worker, Task.Status.AWAITING_PROCESSING)

        # create 3 time entries
        TimeEntry.objects.create(
            worker=self.workers[0],
            date=datetime.datetime.now().date(),
            time_worked=datetime.timedelta(hours=1),
            assignment=assignment)
        TimeEntry.objects.create(
            worker=self.workers[0],
            date=datetime.datetime.now().date(),
            time_worked=datetime.timedelta(hours=1),
            assignment=other_assignment)
        TimeEntry.objects.create(
            worker=self.workers[0],
            date=datetime.datetime.now().date(),
            time_worked=datetime.timedelta(minutes=15),
            assignment=assignment)

        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_ids': [project.id]},
            format='json')
        returned = load_encoded_json(response.content)
        returned_task = returned[str(project.id)]['tasks']
        returned_assignment = returned_task['step1']['assignments'][0]
        recorded_time = returned_assignment['recorded_work_time']
        self.assertEqual(recorded_time, 105*60)  # 1:15 + 0:30

    def test_get_workflow_steps(self):
        # See orchestra.tests.helpers.fixtures for workflow description
        steps = get_workflow_steps('w3', 'crazy_workflow')
        slugs = [step['slug'] for step in steps]

        self.assertTrue(slugs.index('stepC') > slugs.index('stepA'))
        self.assertTrue(slugs.index('stepC') > slugs.index('stepB'))
        self.assertTrue(slugs.index('stepD') > slugs.index('stepA'))
        self.assertTrue(slugs.index('stepD') > slugs.index('stepB'))
        self.assertTrue(slugs.index('stepE') > slugs.index('stepC'))
        self.assertTrue(slugs.index('stepE') > slugs.index('stepD'))
        self.assertTrue(slugs.index('stepF') > slugs.index('stepE'))
        self.assertTrue(slugs.index('stepG') > slugs.index('stepF'))
        self.assertTrue(slugs.index('stepH') > slugs.index('stepF'))

        steps = get_workflow_steps('w4', 'erroneous_workflow_1')

        with self.assertRaises(MalformedDependencyException):
            steps = get_workflow_steps('w5', 'erroneous_workflow_2')

    def test_get_project_information(self):
        projects = Project.objects.all()[:2]
        projects_info = get_project_information(
            [p.pk for p in projects])
        a_project_info_key = list(projects_info.keys())[0]
        a_project_info = projects_info[a_project_info_key]
        self.assertTrue(isinstance(a_project_info['project'], dict))
        self.assertTrue(isinstance(a_project_info['tasks'], dict))
        self.assertTrue(isinstance(a_project_info['steps'], list))

    @patch.object(
        Service, '_create_drive_service',
        new=mock_create_drive_service)
    def test_create_project(self):
        tasks_awaiting_processing = (
            Task.objects
            .filter(status=Task.Status.AWAITING_PROCESSING)
            .count())
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1',
             'workflow_version_slug': 'test_workflow',
             'description': 'short test description',
             'priority': 10,
             'task_class': 'real',
             'project_data': {}},
            format='json')
        self.assertEqual(response.status_code, 200)

        self.assertEqual((Task.objects
                          .filter(status=Task.Status.AWAITING_PROCESSING)
                          .count()),
                         tasks_awaiting_processing + 1)

        # Creating a 'training' project should set task_class correctly.
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1',
             'workflow_version_slug': 'test_workflow',
             'description': 'short test description',
             'priority': 10,
             'task_class': 'training',
             'project_data': {}},
            format='json')
        self.assertEqual(response.status_code, 200)
        project_id = load_encoded_json(response.content)['project_id']
        self.assertEqual(Project.objects.get(id=project_id).task_class,
                         WorkerCertification.TaskClass.TRAINING)

        # Creating a project with missing parameters should fail.
        response = self.api_client.post(
            '/orchestra/api/project/create_project/',
            {'workflow_slug': 'w1'},
            format='json')
        self.ensure_response(response,
                             {'error': 400,
                              'message': 'One of the parameters is missing'},
                             400)

    def test_workflow_types(self):
        response = self.api_client.get(
            '/orchestra/api/project/workflow_types/', format='json')
        self.assertEqual(response.status_code, 200)

        workflows = load_encoded_json(response.content)['workflows']
        workflows = dict(workflows)
        self.assertEqual(
            workflows,
            {
                workflow_slug: {
                    'name': workflow.name,
                    'versions': {
                        v.slug: {
                            'name': v.name,
                            'description': v.description,
                        }
                        for v in workflow.versions.all()
                    }
                }
                for workflow_slug, workflow in self.workflows.items()
            }
        )

    def _make_assign_worker_task_request(self, worker_id,
                                         task_id, success=True):
        response = self.api_client.post(
            '/orchestra/api/project/assign_worker_to_task/',
            {
                'worker_id': worker_id,
                'task_id': task_id,
            },
            format='json')
        self.assertEqual(response.status_code, 200)
        data = load_encoded_json(response.content)
        self.assertEqual(data['success'], success)
        return data

    def test_assign_worker_to_task(self):
        worker = self.workers[0]

        # We should be able to assign to an unassigned task
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(worker.id, task.id)
        self.assertTrue(query.exists())
        query.delete()

        # Nonsense arguments
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(None, None, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Invalid Worker
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(
            -1, task.id, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Invalid Task
        task = self.tasks['awaiting_processing']
        query = TaskAssignment.objects.filter(worker=worker, task=task)
        self.assertFalse(query.exists())
        data = self._make_assign_worker_task_request(
            worker.id, -1, success=False)
        self.assertTrue('error' in data['errors'])
        self.assertFalse(query.exists())

        # Worker0 is not a reviewer so this should fail
        task = self.tasks['review_task']
        data = self._make_assign_worker_task_request(
            worker.id, task.id, success=False)
        self.assertTrue('worker_certification_error' in data['errors'])
        self.assertFalse(query.exists())

        # Can't assign to aborted task
        task = self.tasks['aborted']
        data = self._make_assign_worker_task_request(
            worker.id, task.id, success=False)
        self.assertTrue('task_assignment_error' in data['errors'])
        self.assertFalse(query.exists())

    @patch('orchestra.project_api.views.message_experts_slack_group')
    def test_message_project_team(self, mock_message_slack_group):
        project = ProjectFactory(slack_group_id='test-project-1')
        url = '/orchestra/api/project/message_project_team/'
        test_message = 'this is a test message'
        response = self.api_client.post(
            url,
            {'message': test_message, 'project_id': project.id},
            format='json')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(mock_message_slack_group.called)
        # No project id provided
        response = self.api_client.post(
            url, {'message': test_message}, format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json()['message'],
            ('An object with `message` and `project_id` attributes'
             ' should be supplied'))
        # No message provided
        response = self.api_client.post(
            url, {'project_id': project.id}, format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json()['message'],
            ('An object with `message` and `project_id` attributes'
             ' should be supplied'))
        # Non-existent project_id provided
        response = self.api_client.post(
            url, {'message': 'text', 'project_id': 123456}, format='json')
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json()['message'],
                         'No project for given id')

    def test_permissions(self):
        self.api_client.force_authenticate(user=AnonymousUser())

        response = self.api_client.post(
            '/orchestra/api/project/project_information/',
            {'project_id': 1},
            format='json')
        self.assertEqual(response.status_code, 403)
        returned = load_encoded_json(response.content)
        self.assertEqual(
            returned,
            {'detail': 'You do not have permission to perform this action.'})

    def _delete_keys(self, obj):
        unimportant_keys = (
            'id',
            'task',
            'short_description',
            'start_datetime',
            'end_datetime'
        )
        if isinstance(obj, list):
            for item in obj:
                self._delete_keys(item)

        elif isinstance(obj, dict):
            for key in unimportant_keys:
                try:
                    del obj[key]
                except KeyError:
                    pass
            for value in obj.values():
                self._delete_keys(value)


@override_settings(ORCHESTRA_PROJECT_API_CREDENTIALS={'a': 'b'})
class ProjectAPIAuthTestCase(OrchestraTestCase):

    def setUp(self):
        super().setUp()
        setup_models(self)
        self.authentication = OrchestraProjectAPIAuthentication()

    def test_auth_success(self):
        self.assertEqual(
            self.authentication.fetch_user_data('a'),
            (SignedUser(), 'b'))

    def test_no_key(self):
        with self.assertRaises(AuthenticationFailed):
            self.authentication.fetch_user_data('c')

    def test_misconfigured_credentials(self):
        tmp_setting = settings.ORCHESTRA_PROJECT_API_CREDENTIALS
        del settings.ORCHESTRA_PROJECT_API_CREDENTIALS
        with self.assertRaises(AuthenticationFailed):
            self.authentication.fetch_user_data('a')
        settings.ORCHESTRA_PROJECT_API_CREDENTIALS = tmp_setting
        self.assertEqual(
            self.authentication.fetch_user_data('a'),
            (SignedUser(), 'b'))


class TestTodoApiViewsetTests(EndpointTestCase):
    def setUp(self):
        super().setUp()
        self.request_client = APIClient(enforce_csrf_checks=True)
        self.request_client.force_authenticate(user=SignedUser())
        setup_models(self)
        self.workflow_version = WorkflowVersionFactory()
        self.step = StepFactory(
            slug='step-slug',
            workflow_version=self.workflow_version)
        self.project = ProjectFactory(
            workflow_version=self.workflow_version)
        self.list_url = reverse('orchestra:api:todo-api-list')
        self.todo = TodoFactory(project=self.project)
        self.todo_with_step = TodoFactory(project=self.project, step=self.step)

    def test_permissions(self):
        data = {
            'title': 'Testing title 1',
            'project': self.project.id,
            'step': self.step.slug
        }
        request_client = APIClient(enforce_csrf_checks=True)
        resp = request_client.post(
            self.list_url, data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(
            resp.json()['detail'],
            'Authentication credentials were not provided.')

        # Test if a logged in user cannot access this endpoint
        worker = Worker.objects.get(user__username='test_user_6')
        request_client = APIClient(enforce_csrf_checks=True)
        request_client.login(username=worker.user.username,
                             password='defaultpassword')
        resp = request_client.post(
            self.list_url, data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(resp.status_code, 403)
        self.assertEqual(
            resp.json()['detail'],
            'Authentication credentials were not provided.')

    @patch('orchestra.todos.views.notify_todo_created')
    def test_create(self, mock_notify):
        data = {
            'title': 'Testing create action',
            'project': self.project.id,
            'step': self.step.slug,
            'additional_data': {
                'some_key': 1,
                'other_key': None,
                'some_str': 'test',
                'some_bool': False
            }
        }
        resp = self.request_client.post(
            self.list_url, data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(resp.status_code, 201)
        todos = Todo.objects.filter(
            title__startswith='Testing create action',
            project=self.project,
            step=self.step)
        self.assertEqual(todos.count(), 1)
        self.assertTrue(mock_notify.called)

    def test_bulk_create(self):
        todos = Todo.objects.filter(title__startswith='Testing title ')
        self.assertEqual(len(todos), 0)
        data = [
            {
                'title': 'Testing title {}'.format(x),
                'project': self.project.id,
                'step': self.step.slug,
                'additional_data': {
                    'some_key': 1,
                    'other_key': None,
                    'some_str': 'test',
                    'some_bool': False
                }
            } for x in range(10)
        ]
        resp = self.request_client.post(
            self.list_url, data=json.dumps(data),
            content_type='application/json')
        self.assertEqual(resp.status_code, 201)
        todos = Todo.objects.filter(
            title__startswith='Testing title ',
            project=self.project,
            step=self.step)
        self.assertEqual(len(todos), 10)

    def test_get_single_todo_by_pk(self):
        detail_url = reverse(
            'orchestra:api:todo-api-detail',
            kwargs={'pk': self.todo.id})
        resp = self.request_client.get(detail_url)
        self.assertEqual(resp.status_code, 200)

    def test_get_list_of_todos_with_filters_project_id(self):
        url_with_project_filter = '{}?project__id={}'.format(
            self.list_url, self.project.id)
        resp = self.request_client.get(url_with_project_filter)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 2)

    def test_get_list_of_todos_with_filters_step_slug(self):
        url_with_step_filter = '{}?step__slug={}'.format(
            self.list_url, self.todo_with_step.step.slug)
        resp = self.request_client.get(url_with_step_filter)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 1)
        self.assertEqual(resp.json()[0]['step'], self.todo_with_step.step.slug)

    def test_get_list_of_todos_with_filters_project_id_and_step_slug(self):
        url_with_filters = '{}?project__id={}&step__slug={}'.format(
            self.list_url, self.project.id, self.todo_with_step.step.slug)
        resp = self.request_client.get(url_with_filters)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()[0]['step'], self.todo_with_step.step.slug)

    def test_get_list_of_todos_with_filters_todo_ids(self):
        # Filter by existing todo ids
        ids_to_filter_by = [self.todo.id, self.todo_with_step.id]
        url_with_filters = '{}?&q={}'.format(
            self.list_url,
            json.dumps({'id__in': ids_to_filter_by}))
        resp = self.request_client.get(url_with_filters)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 2)
        for todo in resp.json():
            self.assertTrue(todo['id'] in ids_to_filter_by)

        # Filter by non-existent id
        ids_to_filter_by = [112233445589]
        url_with_filters = '{}?&q={}'.format(
            self.list_url,
            json.dumps({'id__in': ids_to_filter_by}))
        resp = self.request_client.get(url_with_filters)
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(len(resp.json()), 0)

    @patch('orchestra.todos.views.notify_single_todo_update')
    def test_update_functionality(self, mock_notify):
        todo1 = TodoFactory(
            project=self.project, step=self.step, title='Test title1')
        todo2 = TodoFactory(
            project=self.project, step=self.step, title='Test title2')
        # Set title of the todo2 to todo1
        serialized = BulkTodoSerializer(todo2).data
        detail_url = reverse(
            'orchestra:api:todo-api-detail',
            kwargs={'pk': todo1.id})
        resp = self.request_client.put(
            detail_url,
            data=json.dumps(serialized),
            content_type='application/json')
        self.assertEqual(resp.status_code, 200)

        # Check if title is updated
        updated_todo_1 = Todo.objects.get(pk=todo1.pk)
        self.assertEqual(updated_todo_1.title, todo2.title)

    @patch('orchestra.todos.views.notify_single_todo_update')
    def test_partial_update_functionality(self, mock_notify):
        detail_url = reverse(
            'orchestra:api:todo-api-detail',
            kwargs={'pk': self.todo.id})
        expected_title = 'Partial update title'
        resp = self.request_client.patch(
            detail_url,
            data=json.dumps({
                'title': expected_title,
                'step': self.step.slug,
                'project': self.project.id
            }),
            content_type='application/json')
        self.assertEqual(resp.status_code, 200)
        self.assertEqual(resp.json()['title'], expected_title)
        self.assertEqual(resp.json()['step'], self.step.slug)
        self.assertEqual(resp.json()['project'], self.project.id)

        resp = self.request_client.patch(
            detail_url,
            data=json.dumps({
                'title': expected_title,
                'step': self.step.id,
            }),
            content_type='application/json')
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(
            resp.json()['project'],
            ['if step is given, project should also be supplied.'])

    def test_destroy_functionality(self):
        all_todos_count = Todo.objects.count()
        self.assertEqual(all_todos_count, 2)
        detail_url = reverse(
            'orchestra:api:todo-api-detail',
            kwargs={'pk': self.todo.id})
        resp = self.request_client.delete(detail_url)
        self.assertEqual(resp.status_code, 204)
        all_todos_count = Todo.objects.count()
        self.assertEqual(all_todos_count, 1)

        marked_as_deleted = Todo.unsafe_objects.get(pk=self.todo.pk)
        self.assertTrue(marked_as_deleted.is_deleted)
        self.assertEqual(marked_as_deleted, self.todo)

    def test_bulk_update(self):
        todo1 = TodoFactory(
            project=self.project, step=self.step, title='Test title1')
        todo2 = TodoFactory(
            project=self.project, step=self.step, title='Test title2')
        todo3 = TodoFactory(
            project=self.project, step=self.step, title='Test title3')
        todo_wo_project = TodoFactory(
            title='Test title3')
        todo_should_not_be_updated = TodoFactory(
            project=self.project, step=self.step, title='Not updated')
        serialized = BulkTodoSerializer([todo3, todo2, todo1], many=True).data
        # Change titles
        updated = [
            self._change_attr(x, 'title',  'updated title {}'.format(x['id']))
            for x in serialized]
        resp = self.request_client.put(
            self.list_url, data=json.dumps(updated),
            content_type='application/json')
        self.assertEqual(resp.status_code, 200)

        updated_todos = Todo.objects.filter(
            id__in=[todo1.id, todo2.id, todo3.id])
        for todo in updated_todos:
            self.assertEqual(todo.title, 'updated title {}'.format(todo.id))
        self.assertEqual(todo_should_not_be_updated.title, 'Not updated')

        serialized = BulkTodoSerializer([todo_wo_project], many=True).data
        updated = [
            self._change_attr(x, 'title',  'updated title {}'.format(x['id']))
            for x in serialized]
        resp = self.request_client.put(
            self.list_url, data=json.dumps(updated),
            content_type='application/json')
        self.assertEqual(resp.status_code, 400)
        self.assertEqual(resp.json()[0]['step'],
                         ['This field may not be null.'])

    def test_bulk_partial_update(self):
        todo1 = TodoFactory(
            project=self.project, step=self.step, title='Test title1')
        todo2 = TodoFactory(
            project=self.project, step=self.step, title='Test title2')
        todo3 = TodoFactory(
            project=self.project, step=self.step, title='Test title3')
        todo_should_not_be_updated = TodoFactory(
            project=self.project, step=self.step, title='Not updated')
        # Change titles
        todos_with_updated_titles = [{
            'id': x.id,
            'title': 'Updated title {}'.format(x.id),
            'step': x.step.slug,
            'project': x.project.id
        } for x in [todo1, todo3, todo2]]
        resp = self.request_client.patch(
            self.list_url, data=json.dumps(todos_with_updated_titles),
            content_type='application/json')
        self.assertEqual(resp.status_code, 200)

        updated_todos = Todo.objects.filter(
            id__in=[todo1.id, todo2.id, todo3.id])
        for todo in updated_todos:
            self.assertEqual(todo.title, 'Updated title {}'.format(todo.id))
        self.assertEqual(todo_should_not_be_updated.title, 'Not updated')

    def _change_attr(self, item, attr, value):
        item[attr] = value
        return item



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2018-08-21 18:04
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('orchestra', '0080_completed_project_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='step',
            name='todolist_templates_to_apply',
            field=models.ManyToManyField(blank=True, to='orchestra.TodoListTemplate'),
        ),
    ]
